# -*- coding: utf-8 -*-
"""
Created on Wed Jul 20 15:13:47 2022

Deforms a CT or CBCT volume by rotating the head. Output is a deformed dicom volume

@author: mgar5380
"""

#Basics
import sys
import json
import glob
import os
import re
import shutil
import argparse
import numpy as np
import SimpleITK as sitk
from pathlib import Path

#Platipy modules
from platipy.dicom.io.crawl import process_dicom_directory
from platipy.dicom.io.nifti_to_series import convert_nifti_to_dicom_series
from platipy.dicom.io.nifti_to_rtstruct import convert_nifti
from platipy.imaging.generation.mask import get_external_mask
from platipy.imaging.generation.dvf import generate_field_shift
from platipy.imaging.registration.utils import apply_transform

#Custom modules
from DeformHeadCT.deformation import (
    generate_field_rotation,
    PrepareRegistration
    )
from DeformHeadCT.utils import get_head_mask
from DeformHeadCT.DataPreparation import (
    MoveDCMFiles,
    GetPointOfRotation
    )


def DeformationScript(InfoFile,RegParamFile='examples/Elastix_BSpline_OpenCL_RigidPenalty.txt',DeleteTempFile=True,DoRigidNonRigidRegistration=True):#,RegenerateExistingFiles=True):
    """
    Main Section of the code

    Parameters
    ----------
    InfoFile : str
        json file containing the information about the volume
    RegParamFile : str
        Rigid/Non-rigid Parameter file for the registration 
    DeleteTempFile : bool,optional
        If True the temporary files created by this function will be deleted after the new dicom files have been created
    RegenerateExistingFiles : bool,optional
        If False if any of the temporary files (except for the deformed CT and registration files) already exist they won't be recreated.
        If True even if the files already exist they will be regenerated

    Returns
    -------
    None.

    """
    

    #Read json file to get the deformation variables  
    #RegParamFile = ''
    
    with open(InfoFile) as json_file:
        data = json.load(json_file)
    
    assert Path(RegParamFile).is_file(), "Elastix parameter file {} does not exist".format(RegParamFile)
    
    patient_id = data['name']
    
    #input_dcm_dir = data['input_directory'] + '/' + data['CTDir'] #+ '/CT'
    
    if 'axes' in data:
        axes = data['axes']
    else:
        axes = []
    
    if 'angles' in data:
        angles = data['angles']
    else:
        print('No Head Rotations detected, determined by "angles" variable')
        angles = []
    
    if "StructureFile" in data:
        StructDir = data['StructureFile']
    else:
        print('No "StructureFile" variable detected in Input Json File')
        StructDir = ''    
    
    if StructDir:
        if ('Structure_Shift' in data) and ('Structure_Names' in data):
            structShiftFlag = 1
        else:
            structShiftFlag = 0
        
            if 'Structure_Shift' in data:
                print('No Variable "Structure_Names"')
                
            if 'GTVLabel' in data:
                print('No Variable "Structure_Shift"')
    else:
         structShiftFlag = 0   
    
    #Get centre points of rotation
    if 'angles' in data:
        point_of_rotation = GetPointOfRotation(data)
    
    if 'coordinates_cutoff' in data:
        coordinates_cutoff = data['coordinates_cutoff']
    else:
        if 'angles' in data:
            assert 'coordinates_cutoff' in data, "No variable 'coordinates_cutoff' in json file"
        else:
            coordinates_cutoff = []
            
    
    assert "InputDirectory" in data, "No variable 'InputDirectory' in json file"
    
    InputDir = data['InputDirectory']
    
    assert "OutputDirectory" in data, "No variable 'OutputDirectory' in json file"
    
    OutputDir = data['OutputDirectory']
    
    Path.mkdir(Path(OutputDir),parents=False, exist_ok=True)
    
    if "TempDirectory" in data:
        nifti_directory = Path(data['TempDirectory'])
        
    else:
        print('No Temp directory detected in Input Json File')
        nifti_directory = Path(OutputDir + '/nifti')
    
    Path.mkdir(nifti_directory,parents=False, exist_ok=True)
    
    patientunderscore = patient_id.replace('-','_')
    
    #Move input dicom files into one folder in the folder required for processing (Make sure to delete later)
    
    input_dcm_dir = str(nifti_directory) + '/dicom'
    
    Path.mkdir(Path(input_dcm_dir),parents=False, exist_ok=True)
    
    input_dcm_ct_dir = input_dcm_dir + '/ct'
    Path.mkdir(Path(input_dcm_ct_dir),parents=False, exist_ok=True)
    MoveDCMFiles(InputDir,input_dcm_ct_dir)
    
    if StructDir:
        input_dcm_struct_dir = input_dcm_dir + '/rtstruct'
        Path.mkdir(Path(input_dcm_struct_dir),parents=False, exist_ok=True)
        MoveDCMFiles(StructDir,input_dcm_struct_dir)
    
    #Convert from dicom volumes to nifti volumes. Should auto detect if there is or isn't a struct dir
    process_dicom_directory(input_dcm_dir,output_directory=nifti_directory)

    #Define nifti volumes to work with     
    ct_path = next(iter(nifti_directory.glob(f"{patientunderscore}/IMAGES/*CT*.nii.gz")))

    if StructDir:
        # Read in structures
        structure_paths = nifti_directory.glob(f"{patientunderscore}/STRUCTURES/*.nii.gz")
        structures = {re.findall(".*_RTSTRUCT_(.*).nii.gz", p.name)[0]: sitk.ReadImage(str(p)) for p in structure_paths}

    # Read in image    
    image_ct = sitk.ReadImage(str(ct_path))

    # Automatically generate external mask
    struct_external = get_external_mask(image_ct)
    
    #generate head mask using list of coordinates as cutoff points
    head_mask = get_head_mask(struct_external,coordinates_cutoff, image_ct)
    
    #Define variables
    RotationPt = [];    
    Mask_deformed = [];
    dvf_transform = [];
    dvf_field = sitk.Image(image_ct.GetSize(), sitk.sitkVectorFloat64);
    sitk.Image.CopyInformation(dvf_field,image_ct)    
    
    if np.count_nonzero(angles) > 0:
    
        #Apply the rotations 
        try: 
            len(axes[0])
            LoopNum = len(axes)
        except:
            LoopNum = 1
            
        try:
            angles[0]
        except:
            angles = [angles]
    
        for i in range(0,LoopNum):
            try:
                len(axes[i])
                axisLoop = axes[i]
            except:
                axisLoop = axes
            
            try:
                len(point_of_rotation[i])
                RotationPt.append(point_of_rotation[i])            
            except:
                RotationPt.append(point_of_rotation)
            
            Mask_deformed_loop, dvf_transform_loop, dvf_field_loop, RotationTForm = generate_field_rotation(
                image_ct,
                head_mask,
                tuple(RotationPt[i]),
                axis_of_rotation=axisLoop,
                angle=angles[i]*np.pi/180,
                gaussian_smooth=0
                #mask_bend_from_reference_point = coordinates_cutoff
            )
            Mask_deformed.append(Mask_deformed_loop)
            dvf_transform.append(dvf_transform_loop)
            dvf_field = dvf_field + dvf_field_loop        
    
    #Do the rigid shift of the structures 
    if structShiftFlag:
        if np.count_nonzero(data['Structure_Shift']) > 0:
            if np.size(data['Structure_Shift'][0]) > 1:
                for i in range(0,len(data['Structure_Shift'])):
                    
                    if data['Structure_Names'][i] not in structures:
                        print('No Structure {}'.format(data['Structure_Names'][i]))
                    else:
                    
                        Mask_shift, dvf_shift, dvf_field_shift = generate_field_shift(
                            mask_image = structures[data['Structure_Names'][i]],
                            vector_shift = tuple(data['Structure_Shift'][i]),
                            gaussian_smooth = 2
                        )
                        #dvf_transform.append(dvf_shift)
                        
                        dvfLoop = sitk.Cast(dvf_field_shift, sitk.sitkVectorFloat32)  
                        dvfLoop.CopyInformation(dvf_field)                           
                        
                        try:
                            dvf_field = dvf_field + dvf_field_shift
                        except:
                            dvf_field = dvf_field + sitk.Cast(dvfLoop,sitk.sitkVectorFloat64) 
                        dvf_transform.append(dvf_shift)
                        Mask_deformed.append(Mask_shift)
            else:
                if data['Structure_Names'] not in structures:
                    print('No Structure {}'.format(data['Structure_Names']))
                else:
                    Mask_shift, dvf_shift, dvf_field_shift = generate_field_shift(
                        mask_image = structures[data['Structure_Names']],
                        vector_shift = tuple(data['Structure_Shift']),
                        gaussian_smooth = 2
                    )                
                    dvf_transform.append(dvf_shift)
                    Mask_deformed.append(Mask_shift)
                    
                    dvfLoop = sitk.Cast(dvf_field_shift, sitk.sitkVectorFloat32)  
                    dvfLoop.CopyInformation(dvf_field)     
                
                    try:
                        dvf_field = dvf_field + dvf_field_shift
                    except:
                        dvf_field = dvf_field + sitk.Cast(dvfLoop,sitk.sitkVectorFloat64) 
    
    if DoRigidNonRigidRegistration:
    
        if np.count_nonzero(angles) > 0 or structShiftFlag:
        
            InitialTransform = sitk.DisplacementFieldTransform(
                sitk.Cast(dvf_field, sitk.sitkVectorFloat64)
            )
        
            #Deform CT volume using initial transformation
            image_ct_deformed = apply_transform(image_ct, transform=InitialTransform, interpolator=sitk.sitkLinear)
        
            #Do the rigid/non-rigid registration
            dvf_New = PrepareRegistration(image_ct,image_ct_deformed,nifti_directory,RegParamFile)
        else:
            dvf_New = dvf_field

    #Cast dvf prior to transformation
    FinalTransform = sitk.DisplacementFieldTransform(
        sitk.Cast(dvf_New, sitk.sitkVectorFloat64)
    )
    
    #Use DVF to deform CT
    image_ct_deformed2 = apply_transform(image_ct, transform=FinalTransform, interpolator=sitk.sitkLinear)

    #apply dvf field to deformed_structures
    if StructDir:       
        deformed_structures = {}
        for struct in structures:
            #deformed_structures[struct] = apply_field(structures[struct], dvf_transform, structure=True, default_value=0,interp=sitk.sitkLinear)
            deformed_structures[struct] = apply_transform(structures[struct], transform=FinalTransform, default_value=0,interpolator=sitk.sitkLinear)
            

    outputDirCT = Path(OutputDir + '/ct')    
    outputDirCT.mkdir(exist_ok=True, parents=True)
    
    #write deformed ct to dcm
    convert_nifti_to_dicom_series(
        image=image_ct_deformed2,
        reference_dcm=input_dcm_dir + '/ct',
        output_directory=str(outputDirCT)
    )    
    
    if StructDir:        
        #Create a temporary directory to write the structures to
        TempDirStruct = Path(OutputDir + '/OutputStructures')
        TempDirStruct.mkdir(exist_ok=True, parents=True)
        
        #write structure to nifti format 
        for struct in deformed_structures:
            sitk.WriteImage(deformed_structures[struct],str(TempDirStruct / str(struct + ".nii.gz")))
        
        # rtstruct output dir 
        output_dir_dcm_STRUCT = Path(OutputDir + '/rtstruct')  
        output_dir_dcm_STRUCT.mkdir(exist_ok=True, parents=True)
    
        # dictionary containing path of nifti files
        masks = {}
        for m in os.listdir(TempDirStruct):
            name = m.split('.')[0]
            mask_path = str(TempDirStruct / m)
            masks[name] = mask_path    
    
        #convert nifti files to struct file
        convert_nifti(
            dcm_path = str(outputDirCT / "0.dcm"),
            mask_input = masks, 
            output_file = str(output_dir_dcm_STRUCT / "struct.dcm")
        )       
    
    #Delete any temporary files
    if DeleteTempFile:       
        shutil.rmtree(input_dcm_dir)
        shutil.rmtree(nifti_directory)
        if StructDir:
            shutil.rmtree(TempDirStruct)

#DeformationScript('examples/OneAxisRotationAndGTVShift.json','examples/Elastix_BSpline_OpenCL_RigidPenalty.txt',DeleteTempFile=False,RegenerateExistingFiles=True)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Input is a set of CT dicom files. Deforms the volume and outputs a new set of dicom files containing the deformed CT volume.')
    parser.add_argument('InfoFile', type=str, help='File path of the Input Json file containing the deformation information')
    parser.add_argument('RegParamFile',type=str,default='examples/Elastix_BSpline_OpenCL_RigidPenalty.txt',help='The elastix parameter file that controls the rigid/non-rigid registration. Default location is examples/Elastix_BSpline_OpenCL_RigidPenalty.txt ')

    #Changed so that the default is to delete the temporary files to be in line with the main script. 
    parser.add_argument('DontDeleteTempFile',action='store_true', help='If False will delete the temporary files that are created in the Json file variable TempDirectory.')
    parser.add_argument('DoRigidNonRigidRegistration',action='store_true', help='If True will perform the rigid/non-rigid registration as a final smoothing step using the elastix parameter file.')
    
    args = parser.parse_args()
    assert args.InfoFile, "InfoFile {} does not exist".format(args.Info)
    
    DeformationScript(args.InfoFile,args.RegParamFile,DeleteTempFile=not(args.DontDeleteTempFile),DoRigidNonRigidRegistration=args.DoRigidNonRigidRegistration)
    
    #Map command line arguments to function arguments.
    #DeformationScript(*sys.argv[1:])


# -*- coding: utf-8 -*-
"""
Created on Fri Sep 16 13:58:20 2022

@author: mgar5380
"""
#External Modules
import os
import json
import SimpleITK as sitk
from pathlib import Path

#Platipy Modules
from platipy.dicom.io.crawl import process_dicom_directory
from platipy.dicom.io.nifti_to_rtstruct import convert_nifti
from platipy.dicom.io.nifti_to_series import convert_nifti_to_dicom_series

#CTHeadDeformation Modules
from DeformHeadCT.DataPreparation import (
    MoveDCMFiles,
    GetPointOfRotation
    )

"""
TODO
Add more input checking options
Add gaussian smoothing option for both head rotation and GTV shift
"""

class VolumeDeformation:
    
    def __init__(self,InfoFile = '',patient_id = '',patientunderscore = '',axes = [],angles = [],
                 InputDir = '',StructDir = '',OutputDir = '/Output',nifti_directory = '/temp',Structure_Shift = '',
                 Structure_Names = '',coordinates_cutoff = [], VertDict = {}, verbose=True):
        '''

        Parameters
        ----------
        InfoFile : str, optional
            The file location of the .json file containing the information about the volume being deformed. If empty variables need to be 
            manually entered. If InfoFile is not empty all initial variables are extracted from json file. The default is ''.
        patient_id : str, optional
            patient id as described by the dicom file. The default is ''.
        patientunderscore : str, optional
            patient_id but blank spaces are replaced by underscores. If empty the patientunderscore will be genreated using patient_id. 
            The default is ''.
        axes : array, optional
            nx3 array where the nth element is the axis around which the volume will rotate. See wiki for more info The default is [].
        angles : (float | arr), optional
            angles of rotation. If empty no head rotation will be done. The default is [].
        InputDir : str, optional
            Directory of the ct dicom files to be deformed. The default is ''.
        StructDir : str, optional
            Location of the STRUCT dicom file. The default is ''.
        OutputDir : str, optional
            Directory where the dicom values of the deformed CT will be written to. The default is '/Output'.
        nifti_directory : str, optional
            Location of the temp directory where temporary files will be written to. The default is '/temp'.
        Structure_Shift : array, optional
            1x3 array describing the magnitude of the rigid displacement of the structures in "Structure_Names". The default is ''.
        Structure_Names : (str | array), optional
            The structures to be shifted. The default is ''.
        coordinates_cutoff : array, optional
            2x3 array annotating the co-ordinates that seperate the head and the shoulders. See the wiki for more information. The default is [].
        VertDict : dict, optional
            Voxel location of the intervertebral gaps for the CT scan to be deformed. The default is {}.
        verbose : bool, optional
            If True all the warning messages will be displayed. Messages can be useful for debugging. If False no messages displayed.
            The default is True.

        Returns
        -------
        None.

        '''
        if InfoFile:
            if verbose:
                print('Defining volume parameters using input json file')
            
            #Read json file
            with open(InfoFile) as json_file:
                data = json.load(json_file)               
            
            self.patient_id = data['name']
            
            self.patientunderscore = self.patient_id.replace('-','_').upper()
            
            if 'axes' in data:
                self.axes = data['axes']
            else:
                self.axes = ''

            if 'angles' in data:
                self.angles = data['angles']
                if 'coordinates_cutoff' in data:
                    self.coordinates_cutoff = data['coordinates_cutoff']
                else:
                    if verbose:
                        print('No "coordinates_cutoff" variable detected')
                        
                self.point_of_rotation = GetPointOfRotation(data)
            else:
                self.angles = ''
                if verbose:
                    print('No Head Rotations detected, determined by "angles" variable')     
            
            if 'InputDirectory' in data:
                self.InputDir = data['InputDirectory']
            else:
                if verbose:
                    print('No "InputDirectory" variable detected in input json file')
                self.InputDir = -1  
            
            if 'OutputDirectory' in data:
                self.OutputDir = data['OutputDirectory']
            else:
                self.OutputDir = ''
                if verbose:
                    print('No "OutputDirectory" variable detected in input json file')
  
            if 'TempDirectory' in data:
                self.nifti_directory = Path(data['TempDirectory']) 
            else:
                self.nifti_directory = ''
                if verbose:
                    print('No "TempDirectory" variable detected in input json file')
  
            if 'StructureFile' in data:
                self.StructDir = data['StructureFile']
                
                if 'Structure_Names' in data and 'Structure_Shift' in data:
                    self.Structure_Names = data['Structure_Names']
                    self.Structure_Shift = data['Structure_Shift']
                else:
                    if verbose:
                        print('Variable "Structure_Names" or "Structure_Shift" not provided')
                
            else:
                self.StructDir = ''
                if verbose:
                    print('No "StructureFile" variable detected in Input Json File')

        else:
            if verbose:
                print('Defining volume parameters manually')
            
            self.patient_id = patient_id
            
            #If patientunderscore variable not entered then it can be created using patient_id
            if not patientunderscore:
                self.patientunderscore = self.patient_id.replace('-','_')
            else:
                self.patientunderscore = patientunderscore
            
            self.axes = axes
            self.angles = angles
            self.Structure_Shift = Structure_Shift
            self.coordinates_cutoff = coordinates_cutoff
            
            self.InputDir = InputDir
            self.StructDir = StructDir
            self.OutputDir = OutputDir
            self.nifti_directory = Path(nifti_directory)
            self.Structure_Names = Structure_Names
            
            data = {};
            data['axes'] = self.axes
            if 'Oc-C1' in VertDict:
                data['Oc-C1'] = VertDict['Oc-C1']
            if 'C1-C2' in VertDict:
                data['C1-C2'] = VertDict['C1-C2']
            if 'C2-C3' in VertDict:
                data['C2-C3'] = VertDict['C2-C3']
            #Other vertebrae added in case a more complex model gets used in the future but not needed
            if 'C3-C4' in VertDict:
                data['C3-C4'] = VertDict['C3-C4']
            if 'C4-C5' in VertDict:
                data['C4-C5'] = VertDict['C4-C5']
            if 'C5-C6' in VertDict:
                data['C5-C6'] = VertDict['C5-C6']
            if 'C6-C7' in VertDict:
                data['C6-C7'] = VertDict['C6-C7']
            
            self.point_of_rotation = GetPointOfRotation(data)
           
        # Create output and temporary directories if they don't already exist
        Path.mkdir(self.nifti_directory,parents=False, exist_ok=True)
        Path.mkdir(Path(self.OutputDir),parents=False, exist_ok=True)       
        
        
    def PrepareDcmData(self):
        
        """
        Put the dicom files into the format required by platipy, and convert the dicom files to nifti files. 
        Nifti files stored in TempDirectory.
        
        """
        
        #Move input dicom files into one folder            
        input_dcm_dir = str(self.nifti_directory) + '/dicom'
        Path.mkdir(Path(input_dcm_dir),parents=False, exist_ok=True)
        
        #Create dicom/ct
        input_dcm_ct_dir = input_dcm_dir + '/ct'
        Path.mkdir(Path(input_dcm_ct_dir),parents=False, exist_ok=True)
        MoveDCMFiles(self.InputDir,input_dcm_ct_dir)
        
        #If Structure shift is required, move structure information into dicom/rtstruct
        if self.StructDir:
            input_dcm_struct_dir = input_dcm_dir + '/rtstruct'
            Path.mkdir(Path(input_dcm_struct_dir),parents=False, exist_ok=True)
            MoveDCMFiles(self.StructDir,input_dcm_struct_dir)
        
        #Convert from dicom volumes to nifti volumes. Should auto detect if there is or isn't a struct dir
        process_dicom_directory(input_dcm_dir,output_directory=self.nifti_directory,verbose=True)
        
    def WriteVolumesToFile(self,image_ct_deformed,structShiftFlag=0,deformed_structures = {}):
        """
        Write the deformed ct volume and structures to dicom formats in the directory OutputDir

        Parameters
        ----------
        image_ct_deformed : SimpleITK.Image
            The deformed ct image.
        structShiftFlag : bool or int or whatever, optional
            If true the structures in deformed_structures flag will be written. If 0 then only the image_ct_deformed will be written
        deformed_structures : TYPE, optional
            DESCRIPTION. The default is {}.

        Returns
        -------
        None.

        """
        #Define CT variables
        outputDirCT = Path(self.OutputDir + '/ct')    
        outputDirCT.mkdir(exist_ok=True, parents=True)
        
        input_dcm_dir = str(self.nifti_directory) + '/dicom'
                   
        #write deformed ct to dcm
        convert_nifti_to_dicom_series(
            image=image_ct_deformed,
            reference_dcm=input_dcm_dir + '/ct',
            output_directory=str(outputDirCT)
        )                
               
        if structShiftFlag:
            #Create a temporary directory to write the structures to
            TempDirStruct = Path(self.OutputDir + '/OutputStructures')
            TempDirStruct.mkdir(exist_ok=True, parents=True)
            
            #write structure to nifti format 
            for struct in deformed_structures:
                sitk.WriteImage(deformed_structures[struct],str(TempDirStruct / str(struct + ".nii.gz")))
            
            # rtstruct output dir 
            output_dir_dcm_STRUCT = Path(self.OutputDir + '/rtstruct')  
            output_dir_dcm_STRUCT.mkdir(exist_ok=True, parents=True)
        
            # dictionary containing path of nifti files
            masks = {}
            for m in os.listdir(TempDirStruct):
                name = m.split('.')[0]
                mask_path = str(TempDirStruct / m)
                masks[name] = mask_path    
        
            #convert nifti files to struct file
            convert_nifti(
                dcm_path = str(outputDirCT / "0.dcm"),
                mask_input = masks, 
                output_file = str(output_dir_dcm_STRUCT / "struct.dcm")
            )                   



# -*- coding: utf-8 -*-
"""
Created on Wed Jul 20 15:26:22 2022

@author: mgar5380
"""
#External Modules
import os
import re
import SimpleITK as sitk
import numpy as np
from pathlib import Path
from scipy.spatial.transform import Rotation

#Platipy Modules
#from platipy.imaging.registration.registration import apply_field
from platipy.imaging.generation.dvf import generate_field_shift
from platipy.imaging.registration.utils import apply_transform
from platipy.imaging.generation.mask import get_bone_mask
from platipy.imaging.generation.mask import get_external_mask

#Internal Modules
from DeformHeadCT.utils import get_head_mask

"""
TODO
Add option for gaussian smoothing
"""

class HeadDeformation():
    
    def __init__(self,nifti_directory,patientunderscore,structShiftFlag = 0):
        """
        Initialise class

        Parameters
        ----------
        nifti_directory : Path
            Path to temporary directory.
        patientunderscore : str
            patient identifier.
        structShiftFlag : bool, optional
            If 1 (Or True or whatever) the structures will be shifted. The default is 0.

        Returns
        -------
        None.

        """
        #Define nifti volumes to work with     
        ct_path = next(iter(nifti_directory.glob(f"{patientunderscore}/IMAGES/*CT*.nii.gz")))
        
        # Read in image    
        self.image_ct = sitk.ReadImage(str(ct_path))
        
        if structShiftFlag:
            self.structure_paths = nifti_directory.glob(f"{patientunderscore}/STRUCTURES/*.nii.gz")
        else:
            self.structure_paths = ''
    
        #Define dvf_field
        self.dvf_field = sitk.Image(self.image_ct.GetSize(), sitk.sitkVectorFloat64);
        sitk.Image.CopyInformation(self.dvf_field,self.image_ct)    
    
    def ApplyHeadRotation(self,coordinates_cutoff,angles,axes,point_of_rotation):
        """
        Generates a deformation to rotate the head, which is stored in dvf_field

        Parameters
        ----------
        coordinates_cutoff: array
            2x3 array annotating the co-ordinates that seperate the head and the shoulders. See the wiki for more information.
        angles : (float | arr)
            angles of rotation. If empty no head rotation will be done. 
        axes : array
            nx3 array where the nth element is the axis around which the volume will rotate. See wiki for more info.
        point_of_rotation : array
            nx3 array where each nth element is the centre voxel around which the head rotations will occur. .

        Returns
        -------
        Return 0 if no error, returns -1 if there is a problem with the input variables

        """
        if not angles:
            print('"angles" variable cannot be empty')
            return -1
        
        #Get Masks
        # Automatically generate external mask
        struct_external = get_external_mask(self.image_ct)
        
        #generate head mask using list of coordinates as cutoff points
        head_mask = get_head_mask(struct_external,coordinates_cutoff, self.image_ct)            
        
        #Apply the rotations 
        try: 
            len(axes[0])
            LoopNum = len(axes)
        except:
            LoopNum = 1
            
        try:
            angles[0]
        except:
            angles = [angles]
        
        #Define variables. Do they get used? Who knows. But they are living their best life.   
        Mask_deformed = [];
        dvf_transform = [];
        
        for i in range(0,LoopNum):
            if LoopNum > 1:
                axisLoop = axes[i]
                RotationPt = point_of_rotation[i]
            else:
                axisLoop = axes
                RotationPt = point_of_rotation[0]
                
            print(LoopNum)
            print(RotationPt)
                
            Mask_deformed_loop, dvf_transform_loop, dvf_field_loop, RotationTForm = generate_field_rotation(
                self.image_ct,
                head_mask,
                tuple(RotationPt),
                axis_of_rotation=axisLoop,
                angle=angles[i]*np.pi/180,
                gaussian_smooth=0
                #mask_bend_from_reference_point = coordinates_cutoff
            )
            Mask_deformed.append(Mask_deformed_loop)
            dvf_transform.append(dvf_transform_loop)
            self.dvf_field = self.dvf_field + dvf_field_loop   
    
        return 0
    
    def ApplyGTVShift(self,Structure_Shift,Structure_Names):
        """
        Rigidly Shift the structures defined by Structure_Names by the amount described in Structure_Shift

        Parameters
        ----------
        Structure_Shift : array
            1x3 array describing the magnitude of the rigid displacement of the structures in "Structure_Names".
        Structure_Names : (str | array),
            The structures to be shifted.

        Returns
        -------
        None.

        """
        structures = {re.findall(".*_RTSTRUCT_(.*).nii.gz", p.name)[0]: sitk.ReadImage(str(p)) for p in self.structure_paths}
        
        if np.size(Structure_Shift[0]) > 1:
            for i in range(0,len(Structure_Shift)):
                
                if Structure_Names[i] not in structures:
                    print('No Structure {}'.format(Structure_Names[i]))
                else:
                
                    Mask_shift, dvf_shift, dvf_field_shift = generate_field_shift(
                        mask_image = structures[Structure_Names[i]],
                        vector_shift = tuple(Structure_Shift[i]),
                        gaussian_smooth = 2
                    )
                    #dvf_transform.append(dvf_shift)
                    
                    dvfLoop = sitk.Cast(dvf_field_shift, sitk.sitkVectorFloat32)  
                    dvfLoop.CopyInformation(self.dvf_field)                           
                    
                    try:
                        self.dvf_field = self.dvf_field + dvf_field_shift
                    except:
                        self.dvf_field = self.dvf_field + sitk.Cast(dvfLoop,sitk.sitkVectorFloat64) 
                    #dvf_transform.append(dvf_shift)
                    #Mask_deformed.append(Mask_shift)
        else:
            if Structure_Names not in structures:
                print('No Structure {}'.format(Structure_Names))
            else:
                Mask_shift, dvf_shift, dvf_field_shift = generate_field_shift(
                    mask_image = structures[Structure_Names],
                    vector_shift = tuple(Structure_Shift),
                    gaussian_smooth = 2
                )                
                #dvf_transform.append(dvf_shift)
                #Mask_deformed.append(Mask_shift)
                
                dvfLoop = sitk.Cast(dvf_field_shift, sitk.sitkVectorFloat32)  
                dvfLoop.CopyInformation(self.dvf_field)     
            
                try:
                    self.dvf_field = self.dvf_field + dvf_field_shift
                except:
                    self.dvf_field = self.dvf_field + sitk.Cast(dvfLoop,sitk.sitkVectorFloat64) 




def generate_field_rotation(
    reference_image,
    body_mask,
    reference_point,
    axis_of_rotation=[0, 0, -1],
    angle=0.1,
    mask_bend_from_reference_point=("z", "inf"),
    gaussian_smooth=5,
):
    """
    Generates a synthetic field characterised by specified rotation
    Typically, this field would be used to simulate a moving head and so masking is important.
    Args:
        reference_image ([SimpleITK.Image]): The image to be deformed.
        body_mask ([SimpleITK.Image]): A binary mask in which the deformation field will be defined
        reference_point ([tuple]): The point (z,y,x) about which the rotation field is defined.
        axis_of_rotation (tuple, optional): The axis of rotation (z,y,x). Defaults to [0, 0, -1].
        angle (double, optional): The deformation vector length at each point will equal scale multiplied by the distance to that point from reference_point. Defaults to 1.
        mask_bend_from_reference_point (tuple, optional): The dimension (z=axial, y=coronal, x=sagittal) and limit (inf/sup, post/ant, left/right) for masking the vector field, relative to the reference point. Defaults to ("z", "inf").
        gaussian_smooth (int | list, optional): Scale of a Gaussian kernel used to smooth the deformation vector field. Defaults to 5.
    Returns:
        [SimpleITK.Image]: The binary mask following the expansion.
        [SimpleITK.DisplacementFieldTransform]: The transform representing the expansion.
        [SimpleITK.Image]: The displacement vector field representing the expansion.
    """
    body_mask_arr = sitk.GetArrayFromImage(body_mask)

    pt_arr = np.array(np.where(body_mask_arr))
    vector_ref_to_pt = pt_arr - np.array(reference_point)[:, None]
    
    vector_ref_to_pt=vector_ref_to_pt[::-1].T;
    axis_of_rotation = np.array(axis_of_rotation[::-1])
    axis_of_rotation = axis_of_rotation / np.linalg.norm(axis_of_rotation)
    
    # initialise scipy rotation instance
    rotation = Rotation.from_rotvec(-1*angle * axis_of_rotation)
    #ector_ref_to_pt = np.ndarray.transpose(vector_ref_to_pt)
    rotated_vector_ref_to_pt = rotation.apply(vector_ref_to_pt)
    deformation_vectors = rotated_vector_ref_to_pt - vector_ref_to_pt;


    dvf_template = sitk.Image(reference_image.GetSize(), sitk.sitkVectorFloat64, 3)
    dvf_template_arr = sitk.GetArrayFromImage(dvf_template)
    
    dvf_template_arr[np.where(body_mask_arr)] = deformation_vectors;
#     if scale is not False:
#         dvf_template_arr[np.where(body_mask_arr)] = deformation_vectors * scale

    dvf_template = sitk.GetImageFromArray(dvf_template_arr)
    dvf_template.CopyInformation(reference_image)

    # smooth
    if np.any(gaussian_smooth):

        if not hasattr(gaussian_smooth, "__iter__"):
            gaussian_smooth = (gaussian_smooth,) * 3

        dvf_template = sitk.SmoothingRecursiveGaussian(dvf_template, gaussian_smooth)

    dvf_tfm = sitk.DisplacementFieldTransform(
        sitk.Cast(dvf_template, sitk.sitkVectorFloat64)
    )

    reference_image_bend = apply_transform(
        reference_image,
        transform=dvf_tfm,
        default_value=int(sitk.GetArrayViewFromImage(reference_image).min()),
        interpolator=2,
    )
    return reference_image_bend, dvf_tfm, dvf_template, rotation

def PrepareRegistration(image_ct,image_ct_deformed,TempDir,paramFilePath):
    """
    Prepare the files and do the rigid/non-rigid registration with files stored in a temp directory

    Parameters
    ----------
    image_ct : SimpleITK.Image
        Original undeformed ct image.
    image_ct_deformed : SimpleITK.Image
        Deformed ct image.
    TempDir : Path
        Path where the temporarily files will be stored.
    paramFilePath : str
        FileName of the registration parameter.

    Returns
    -------
    dvf_New : SimpleITK.Image
        DVF output of the registration.

    """
    
    #Prepare data for rigid/non-rigid registration
    #Normalise and write initial and deformed CT vols
    image_ct_norm = sitk.Normalize(image_ct)
    sitk.WriteImage(image_ct_norm,str(TempDir) + '/CTNorm.nii.gz')
    
    image_ct_deformed_norm = sitk.Normalize(image_ct_deformed)
    sitk.WriteImage(image_ct_deformed_norm,str(TempDir) + '/CTWarpedNorm.nii.gz')
    
    #Create Bone Masks
    BoneMaskInit = get_bone_mask(image_ct)
    BoneMaskWarped = get_bone_mask(image_ct_deformed)
    BoneMaskInitNorm = sitk.Normalize(BoneMaskInit)
    BoneMaskWarpedNorm = sitk.Normalize(BoneMaskWarped)
    sitk.WriteImage(BoneMaskInitNorm,str(TempDir) + '/BoneMaskInit.nii.gz')
    sitk.WriteImage(BoneMaskWarpedNorm,str(TempDir) + '/BoneMaskWarped.nii.gz')
    
    #Do registration
    FixedImage = str(TempDir) +  '/CTWarpedNorm.nii.gz'
    MovingImage = str(TempDir) +  '/CTNorm.nii.gz'
    #parameterFilePath = intermediate_dir + '/Elastix_BSpline_OpenCL_RigidPenalty.txt'

    outputPath = str(TempDir)    

    ParamFileFinal = ModifyElastixParameterFile(paramFilePath,TempDir)
    
    os.system('elastix -f "{}" -m "{}" -p "{}" -out "{}"'.format(FixedImage,MovingImage,ParamFileFinal,outputPath))
    
    ParamFile = outputPath + '/TransformParameters.0.txt'
    
    os.system('transformix -def all -tp "{}" -out "{}"'.format(ParamFile, outputPath))

    dvf_New = sitk.ReadImage(outputPath + '/deformationField.mha')    
    
    return dvf_New

def ModifyElastixParameterFile(paramFilePath,TempDir):
    """
    Modify the elastix parameter file so include the recently create bone mask file

    Parameters
    ----------
    paramFilePath : str
        Name of the input elastix parameter file.
    TempDir : Path
        Path where the temporarily files will be stored.

    Returns
    -------
    paramFileOutput: str
        File name of the parameter file to be used for the registration with the correct (hopefully) bonemask file

    """
    
    with open(paramFilePath) as f:
        lines = f.readlines()
    
    for i,line in enumerate(lines):
        if 'MovingRigidityImageName' in line:
            line = '(MovingRigidityImageName "' + str(TempDir) + '/BoneMaskInit.nii.gz' +'")'
            lines[i] = line
            
    paramFileOutput = str(TempDir) + '/ModifiedElastixParameter.txt'
    
    with open(paramFileOutput, 'w') as f:
        for line in lines:
            f.write(line)
            #f.write('\n') 
    
    return paramFileOutput


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 20 15:15:50 2022

@author: mgar5380
"""
import SimpleITK as sitk
import numpy as np

def CombineDVFs(DVFUnSmoothedImg,DVFSmoothedImg,coordinates_cutoff,RotationPt):
    """
    Combines the smoothed and unsmoothed DVFs. Points above the line between 
    the rotation point and cutoff point are smoothed and points below the line 
    are the unsmoothed DVF
    
    Parameters
    ----------
    DVFUnSmoothedImg : SimpleITK Image
        DESCRIPTION.
    DVFSmoothedImg : SimpleITK Image
        DESCRIPTION.
    coordinates_cutoff : numpy array
        DESCRIPTION.
    RotationPt : numpy array
        DESCRIPTION.

    Returns
    -------
    DVFNew : SimpleITK Image
        DESCRIPTION.

    """
    
    DVFUnSmoothed = sitk.GetArrayFromImage(DVFUnSmoothedImg)
    DVFSmoothed = sitk.GetArrayFromImage(DVFSmoothedImg)
    #ImageCTArray = sitk.GetArrayFromImage(image_ct)
    ImageMask = np.zeros(DVFSmoothed.shape)
    
    m1 = (RotationPt[-1][0] - coordinates_cutoff[0][0])/(RotationPt[-1][1] - coordinates_cutoff[0][1])
    m2 = (RotationPt[-1][0] - coordinates_cutoff[1][0])/(RotationPt[-1][1] - coordinates_cutoff[1][1])
    m3 = (coordinates_cutoff[0][0] - coordinates_cutoff[1][0])/(coordinates_cutoff[0][1] - coordinates_cutoff[1][1])
    
    b1 = coordinates_cutoff[0][0] - m1*coordinates_cutoff[0][1]
    b2 = coordinates_cutoff[1][0] - m2*coordinates_cutoff[1][1]
    b3 = coordinates_cutoff[1][0] - m3*coordinates_cutoff[1][1]    
    
    for i in range(0,ImageMask.shape[0]):
        for j in range(0,ImageMask.shape[1]):
            if i < m1*j + b1 and i < m2*j + b2: #and i > m3*j + b3:
                ImageMask[i,j,:,:] = DVFSmoothed[i,j,:,:]
            else:
                ImageMask[i,j,:,:] = DVFUnSmoothed[i,j,:,:]   

    DVFNew = sitk.GetImageFromArray(ImageMask)    
    DVFNew.CopyInformation(DVFSmoothedImg)
    return DVFNew

def RotateAroundPoint(RTform,RotationPt,RotationOrigin):
    """
    

    Parameters
    ----------
    RTform : TYPE
        DESCRIPTION.
    RotationPt : TYPE
        DESCRIPTION.
    RotationOrigin : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    RotationVector = np.array(RotationPt) - np.array(RotationOrigin)
    NewPts = RTform.apply(RotationVector) + RotationOrigin
    return NewPts.tolist()
    pass

def get_head_mask(body_mask,list_coordinates, reference_image):
    """
    Generates a binary mask of the patient head contour using the given list of coordinates as the cutoff. Assumes x value is the same for all coordinates in the list for interpolation of z-cutoffs. 
    Args:
        body_mask ([SimpleITK.Image]): A binary mask of the patient external contour.
        list_coordinates (list): 2-D list of coordinates to interpolate boundary for head mask
    Returns:
        [SimpleITK.Image]: The binary mask from cropping using input coordinates.
    """    
    
    body_mask_arr = sitk.GetArrayFromImage(body_mask)
    
    #y and z coordinates 
    y_val = []
    z_val = []
    for coordinate in list_coordinates:
        #print(coordinate)
        y_val.append(coordinate[1])
        z_val.append(coordinate[0])
    y_val = np.array(y_val)
    z_val = np.array(z_val)
    #sort so that y is ascending
    sort_index = np.argsort(y_val)
    y_val = y_val[sort_index]
    ycoord = list(range(0,np.shape(body_mask_arr)[1]))
    
    #use linear interp to obtain z-coord
    zinterp = np.interp(ycoord, y_val, z_val)
    
    #round to integer value 
    zinterp = np.around(zinterp).astype(int)
    
    #set values in binary mask to 0 below the head
    for i in ycoord:
        # np.shape(body_mask_arr)[1]-1-i
        body_mask_arr[: zinterp[i], np.shape(body_mask_arr)[1]-1-i, :] = 0 #filtering points up to reference from above to be excluded
    
    body_mask_new = sitk.GetImageFromArray(body_mask_arr)
    body_mask_new.CopyInformation(body_mask)
    
    return body_mask_new


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 19 16:14:11 2022

@author: mgar5380
"""

#External Modules
import re
import shutil
import argparse
import numpy as np
import SimpleITK as sitk
from pathlib import Path

#Platipy Modules
from platipy.imaging.registration.utils import apply_transform

from DeformHeadCT.VolumeInfo import VolumeDeformation
#Custom Mdules
from DeformHeadCT.deformation import (
    PrepareRegistration,
    HeadDeformation
    )

def DeformationScript(InfoFile,RegParamFile='examples/Elastix_BSpline_OpenCL_RigidPenalty.txt',DeleteTempFile=True,DoRigidNonRigidRegistration=True):#,RegenerateExistingFiles=True):
    """
    Main Section of the code

    Parameters
    ----------
    InfoFile : str
        json file containing the information about the volume
    RegParamFile : str
        Rigid/Non-rigid Parameter file for the registration 
    DeleteTempFile : bool,optional
        If True the temporary files created by this function will be deleted after the new dicom files have been created
    RegenerateExistingFiles : bool,optional
        If False if any of the temporary files (except for the deformed CT and registration files) already exist they won't be recreated.
        If True even if the files already exist they will be regenerated

    Returns
    -------
    None.

    """
    
    assert Path(RegParamFile).is_file(), "Elastix parameter file {} does not exist".format(RegParamFile)

    #Get volume information from json file
    VolInfo = VolumeDeformation(InfoFile=InfoFile)
    
    if VolInfo.StructDir:
        structShiftFlag = 1
    else:
        structShiftFlag = 0

    #Oorganise the data and convert dicom volumes to nifti files
    VolInfo.PrepareDcmData()

    #Do the Head Rotation
    HeadDef = HeadDeformation(VolInfo.nifti_directory,VolInfo.patientunderscore,structShiftFlag)

    if np.count_nonzero(VolInfo.angles) > 0:
        assert HeadDef.ApplyHeadRotation(VolInfo.coordinates_cutoff,VolInfo.angles,VolInfo.axes,VolInfo.point_of_rotation) >= 0   

    #Do the rigid shift of the structures
    if structShiftFlag:
        HeadDef.ApplyGTVShift(VolInfo.Structure_Shift,VolInfo.Structure_Names)

    
    if DoRigidNonRigidRegistration:
    
        if np.count_nonzero(VolInfo.angles) > 0 or structShiftFlag:
        
            InitialTransform = sitk.DisplacementFieldTransform(
                sitk.Cast(HeadDef.dvf_field, sitk.sitkVectorFloat64)
            )
        
            #Deform CT volume using initial transformation
            image_ct_deformed = apply_transform(HeadDef.image_ct, transform=InitialTransform, interpolator=sitk.sitkLinear)
        
            #Do the rigid/non-rigid registration
            dvf_New = PrepareRegistration(HeadDef.image_ct,image_ct_deformed,VolInfo.nifti_directory,RegParamFile)
        else:
            dvf_New = HeadDef.dvf_field

    #Cast dvf prior to transformation
    FinalTransform = sitk.DisplacementFieldTransform(
        sitk.Cast(dvf_New, sitk.sitkVectorFloat64)
    )
    
    #Use DVF to deform CT
    image_ct_deformed2 = apply_transform(HeadDef.image_ct, transform=FinalTransform, interpolator=sitk.sitkLinear)

    deformed_structures = {}

    #apply dvf field to deformed_structures
    if structShiftFlag:       
        
        structures = {re.findall(".*_RTSTRUCT_(.*).nii.gz", p.name)[0]: sitk.ReadImage(str(p)) for p in HeadDef.structure_paths}
        for struct in structures:
            #deformed_structures[struct] = apply_field(structures[struct], dvf_transform, structure=True, default_value=0,interp=sitk.sitkLinear)
            deformed_structures[struct] = apply_transform(structures[struct], transform=FinalTransform, default_value=0,interpolator=sitk.sitkLinear)
        
    #Write the deformed ct image and deformed structures to the OutputDir directory
    VolInfo.WriteVolumesToFile(image_ct_deformed2,structShiftFlag,deformed_structures)

    #Delete any temporary files
    if DeleteTempFile:       
        shutil.rmtree(str(VolInfo.nifti_directory) + '/dicom')
        shutil.rmtree(VolInfo.nifti_directory)
        if structShiftFlag:
            shutil.rmtree(Path(VolInfo.OutputDir + '/OutputStructures'))

#DeformationScript('examples/OneAxisRotationAndGTVShift.json','examples/Elastix_BSpline_OpenCL_RigidPenalty.txt',DeleteTempFile=False,RegenerateExistingFiles=True)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Input is a set of CT dicom files. Deforms the volume and outputs a new set of dicom files containing the deformed CT volume.')
    parser.add_argument('InfoFile', type=str, help='File path of the Input Json file containing the deformation information')
    parser.add_argument('RegParamFile',type=str,default='examples/Elastix_BSpline_OpenCL_RigidPenalty.txt',help='The elastix parameter file that controls the rigid/non-rigid registration. Default location is examples/Elastix_BSpline_OpenCL_RigidPenalty.txt ')

    #Changed so that the default is to delete the temporary files to be in line with the main script. 
    parser.add_argument('DontDeleteTempFile',action='store_true', help='If False will delete the temporary files that are created in the Json file variable TempDirectory.')
    parser.add_argument('DoRigidNonRigidRegistration',action='store_true', help='If True will perform the rigid/non-rigid registration as a final smoothing step using the elastix parameter file.')
    
    args = parser.parse_args()
    assert args.InfoFile, "InfoFile {} does not exist".format(args.Info)
    
    DeformationScript(args.InfoFile,args.RegParamFile,DeleteTempFile=not(args.DontDeleteTempFile),DoRigidNonRigidRegistration=args.DoRigidNonRigidRegistration)
    
    #Map command line arguments to function arguments.
    #DeformationScript(*sys.argv[1:])


# -*- coding: utf-8 -*-

from setuptools import setup
import re, io

__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',  # It excludes inline comment too
    io.open('DeformHeadCT/__init__.py', encoding='utf_8_sig').read()
    ).group(1)
setup(
    name ='CTHeadDeformation',
    python_requires = '>=3.7',
    packages = ['DeformHeadCT'],
    version = __version__,
    license = 'MIT',
    description = 'Perform Realistic Head Deformations',
    author = 'Mark Gardner',
    author_email='mark.gardner@sydney.edu.au',
    url = 'https://github.com/ACRF-Image-X-Institute/CTHeadDeformation',
    download_url = '',
    keywords = ['CT', 'deformation', 'head and neck'],
    install_requires=['SimpleITK', 
                      'platipy'],
    long_description_content_type = 'text/markdown',
    long_description=open('README.md').read(),
)




# -*- coding: utf-8 -*-
"""
Created on Thu Jul 21 11:35:06 2022

@author: mgar5380
"""

import os
import glob
from shutil import copy2
from pathlib import Path

def MoveDCMFiles(Input_dcm_dir,Output_dcm_dir):
    """
    Copies the dicom files across from Input_dcm_dir to Output_dcm_dir, and 
    makes sure the files are in the correct format (.dcm extension) 

    Parameters
    ----------
    Input_dcm_dir : str
        Directory where the initial dicom files are stored.
    Output_dcm_dir : str
        Directory where the dicom files will be written to.

    Returns
    -------
    None.

    """
    
    if Path(Input_dcm_dir).is_dir():    
        dcmList = glob.glob(Input_dcm_dir + '/*')
        print('Copying {} files across to directory'.format(len(dcmList)))
    elif os.path.exists(Path(Input_dcm_dir)):
        dcmList = []
        dcmList.append(Input_dcm_dir)
    else:
        print('File does not exist')
        print(Input_dcm_dir)
    
    assert dcmList, "Input directory is empty"   
    
    #Copy dicom files across to new directory
    for dcmfile in dcmList:
        LoopFileName,LoopFileExt = os.path.splitext(dcmfile)
        #platipy only works when dicom files have the .dcm extension so add the extension if they don't have it. 
        name = os.path.split(dcmfile)[-1]
        if LoopFileExt == '.dcm':            
            copy2(dcmfile,Output_dcm_dir + '/' + name)
        else:
            copy2(dcmfile,Output_dcm_dir + '/' + name + '.dcm')
            
            #os.rename(LoopFileName+LoopFileExt,LoopFileName+LoopFileExt+'.dcm')
            #dcmfile = LoopFileName+LoopFileExt+'.dcm'
        #name = os.path.split(dcmfile)[-1]
        #copy2(Input_dcm_dir + '/' + name,Output_dcm_dir + '/' + name)

def GetPointOfRotation(JsonData):
    """
    Input is the data from the input json data file and the output is the series 
    of 3D points used as the centre of rotation for the head transformations

    Parameters
    ----------
    JsonData : dict
        Deformation and Volume information.

    Returns
    -------
    PointOfRotation : list
        3D points that are the centre of the head rotations. For single rotation 
        list is of size 1x3, for multiple rotations the list is size nx3

    """
    PointOfRotation = []
    try:
        len(JsonData['axes'][0])
        LoopNum = len(JsonData['axes'])
    except:
        LoopNum = 1
            
    
    for i in range(0,LoopNum):
        try:
            len(JsonData['axes'][i])
            AxisArr = JsonData['axes'][i]
        except:
            AxisArr = JsonData['axes']
    
        if abs(AxisArr[0]) == 1:
            RotationPts = JsonData['C1-C2']            
        elif abs(AxisArr[1]) == 1:
            RotationPts = JsonData['C2-C3']
        elif abs(AxisArr[2]) == 1:
            RotationPts = JsonData['Oc-C1']
    
        PointOfRotation.append(RotationPts)
            
    return PointOfRotation